#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Thu Jul 29, 2021 at 02:36 AM +0200

import numpy as np

from .io import read_branches


# Find total number of events (unique events) out of total number of candidates.
def extract_uid(ntp, tree, run_branch='runNumber', event_branch='eventNumber',
                conditional=None, run_array=None, event_array=None):
    if run_array is None or event_array is None:
        run, event = read_branches(ntp, tree, (run_branch, event_branch))
    else:
        run, event = run_array, event_array

    if conditional is not None:
        run = run[conditional]
        event = event[conditional]

    run = np.char.mod('%d', run)
    event = np.char.mod('%d', event)

    run = np.char.add(run, '-')
    ids = np.char.add(run, event)

    uid, idx, count = np.unique(ids, return_index=True, return_counts=True)

    num_of_evt = ids.size
    num_of_ids = uid.size
    num_of_dupl_ids = uid[count > 1].size
    # num_of_evt_w_dupl_id = np.sum(count[count > 1]) - num_of_dupl_ids
    num_of_evt_w_dupl_id = num_of_evt - num_of_ids

    return uid, idx, num_of_evt, num_of_ids, \
        num_of_dupl_ids, num_of_evt_w_dupl_id


def find_common_uid(ntp1, ntp2, tree1, tree2, **kwargs):
    uid1, idx1 = extract_uid(ntp1, tree1, **kwargs)[0:2]
    uid2, idx2 = extract_uid(ntp2, tree2, **kwargs)[0:2]
    uid_comm, uid_comm_idx1, uid_comm_idx2 = np.intersect1d(
        uid1, uid2, assume_unique=True, return_indices=True)

    return uid_comm, idx1[uid_comm_idx1], idx2[uid_comm_idx2]


def gen_histo(array, bins=200, scale=1.05, data_range=None, **kwargs):
    if data_range is None:
        data_min = array.min()
        data_max = array.max()

        data_min = data_min*scale if data_min < 0 else data_min/scale
        data_max = data_max/scale if data_max < 0 else data_max*scale

        return np.histogram(array, bins, (data_min, data_max), **kwargs)

    return np.histogram(array, bins, data_range, **kwargs)



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Sun Jun 13, 2021 at 06:30 PM +0200

import sys
from argparse import Action, ArgumentParser


###########################
# Argument parser actions #
###########################

class DataRangeAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            setattr(namespace, self.dest, [])

        if len(values) % 2 != 0:
            print('Odd number of min, max pairs!')
            sys.exit(255)

        values = [float(v) for v in values]
        min_max_pairs = self.divide_list_in_chunk(values)
        setattr(namespace, self.dest, min_max_pairs)

    @staticmethod
    def divide_list_in_chunk(lst, chunk_size=2):
        return [lst[i:i + chunk_size] for i in range(0, len(lst), chunk_size)]


class DataPairAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            setattr(namespace, self.dest, [])

        try:
            val_min, val_max = [float(i) for i in values]
            setattr(namespace, self.dest, [val_min, val_max])
        except ValueError:
            print('Invalid value for pair: {}'.format(values))


##############################
# Bare-bone argument parsers #
##############################

def single_ntuple_parser_no_output(descr, parser=None):
    parser = ArgumentParser(description=descr) if parser is None else parser

    parser.add_argument('-n', '--ref',
                        required=True,
                        help='''
path to reference ntuple.''')

    parser.add_argument('-t', '--ref-tree',
                        required=True,
                        help='''
tree name for the reference ntuple.''')

    parser.add_argument('--bins',
                        type=int,
                        default=25,
                        help='''
specify binning.''')

    return parser


def double_ntuple_parser_no_output(descr, parser=None):
    parser = single_ntuple_parser_no_output(descr) if parser is None else parser

    parser.add_argument('-N', '--comp',
                        required=True,
                        help='''
path to comparison ntuple.''')

    parser.add_argument('-T', '--comp-tree',
                        required=True,
                        help='''
tree name for the comparison ntuple.''')

    return parser


def single_branch_parser_no_output(descr, parser=None):
    parser = ArgumentParser(description=descr) if parser is None else parser

    parser.add_argument('-n', '--ref',
                        required=True,
                        action='append',
                        help='''
paths to reference ntuple-tree combos.''')

    parser.add_argument('-b', '--ref-branch',
                        required=True,
                        action='append',
                        nargs='+',
                        help='''
branch names in reference trees.''')

    parser.add_argument('--bins',
                        type=int,
                        default=25,
                        help='''
specify binning.''')

    return parser


def diff_branch_parser_no_output(descr, parser=None):
    parser = single_branch_parser_no_output(descr) if parser is None else parser

    parser.add_argument('-N', '--comp',
                        required=True,
                        action='append',
                        help='''
paths to comparison ntuple-tree combos.''')

    parser.add_argument('-B', '--comp-branch',
                        required=True,
                        action='append',
                        nargs='+',
                        help='''
branch names in comparison trees.''')

    return parser


###########
# Helpers #
###########

def split_ntp_tree(ntp_tree):
    ntp, tree = ntp_tree.split('.root/')
    ntp += '.root'
    return ntp, tree



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Mon Jun 14, 2021 at 03:35 AM +0200

import unittest

from context import pyTuplingUtils as ptu
from context import pwd


class FilterKwargsFuncTest(unittest.TestCase):
    @staticmethod
    def sample_func(a=1, b=2):
        pass

    def test_args(self):
        kwargs = {'a': 2, 'c': 3, 'd': 4}
        kw_known, kw_rest = ptu.plot.filter_kwargs_func(
            kwargs, self.sample_func)

        assert kw_known == {'a': 2}
        assert kw_rest == {'c': 3, 'd': 4}


class AxStyleTest(unittest.TestCase):
    def test_ax_errorbar_all_known_kw(self):
        assert ptu.plot.ax_add_args_errorbar('a', 'black', marker='x') == {
            'label': 'a',
            'ls': 'none',
            'color': 'black',
            'marker': 'x',
            'markeredgecolor': 'none',
            'yerr': None
        }

    def test_ax_errorbar_some_unknown_kw(self):
        assert ptu.plot.ax_add_args_errorbar(
            'a', 'black', marker='x', test='test') == {
            'label': 'a',
            'ls': 'none',
            'color': 'black',
            'marker': 'x',
            'markeredgecolor': 'none',
            'yerr': None,
            'test': 'test'
        }



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Mon May 24, 2021 at 10:06 PM +0200

import uproot

from dataclasses import dataclass
from typing import Union, Optional
from numpy import sum
from numpy import logical_and as AND
from copy import deepcopy

from pyTuplingUtils.boolean.eval import BooleanEvaluator
from pyTuplingUtils.utils import extract_uid
from pyTuplingUtils.io import read_branches


def cutflow_uniq_events_outer(
        ntp, tree, run_branch='runNumber', event_branch='eventNumber'):
    run, event = read_branches(ntp, tree, (run_branch, event_branch))

    def inner(ntp_i, tree_i, arr):
        return extract_uid(ntp_i, tree_i, conditional=arr,
                           run_array=run, event_array=event)[3]

    return inner


@dataclass
class CutflowRule:
    cond: str = 'true'
    name: Optional[str] = None
    compare_to: Union[str, int] = 'r:-1'
    explicit: bool = False
    key: Optional[str] = None


class CutflowGen:
    def __init__(self, ntp_path, tree, rules, init_num, **kwargs):
        self.rules = self.strip_multiline_str(rules)
        self.init_num = init_num

        self.ntp = uproot.open(ntp_path)
        self.tree = tree
        self.exe = BooleanEvaluator(self.ntp, tree, **kwargs)

        self.debug_raw_output = []

    def do(self, output_regulator=lambda ntp, tree, arr: sum(arr)):
        ref = {}
        result = {}

        for idx, r in enumerate(self.rules):
            prev_idx = self.find_idx(idx, r.compare_to)

            # If the 'output' entry does not exist, use the default initial
            # number of events/candidates.
            try:
                prev_output = ref[self.rules[prev_idx].cond]['output']
                prev_raw_output = ref[self.rules[prev_idx].cond]['raw_output']
            except Exception:
                prev_output = self.init_num
                prev_raw_output = True

            # Note that 'raw_output' is an array of boolean
            raw_output = self.exe.eval(r.cond)
            if not r.explicit:
                raw_output = AND(prev_raw_output, raw_output)
            self.debug_raw_output.append(raw_output)

            # Here, 'output' is a number
            if True in raw_output:
                output = output_regulator(self.ntp, self.tree, raw_output)
            else:
                output = 0

            cut_result = {'input': prev_output, 'output': output}

            if r.name:
                cut_result['name'] = r.name

            if r.key:
                result[r.key] = cut_result
            else:
                result[r.cond] = cut_result

            # Include the raw array of boolean in the reference.
            ref_cut_result = deepcopy(cut_result)
            ref_cut_result['raw_output'] = raw_output
            ref[r.cond] = ref_cut_result

        return result

    @staticmethod
    def find_idx(ref_idx, raw_idx):
        if isinstance(raw_idx, str):  # relative index
            _, idx = raw_idx.split(':')
            idx = ref_idx + int(idx)
        else:
            idx = int(raw_idx)

        return idx

    @staticmethod
    def strip_multiline_str(rules):
        result = []
        for r in rules:
            stripped_rule = CutflowRule(
                r.cond.replace('\n', ' '),
                r.name, r.compare_to, r.explicit, r.key
            )
            result.append(stripped_rule)
        return result



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Mon May 31, 2021 at 02:55 AM +0200

import unittest
import uproot
import os.path as osp
import numpy as np

from context import pyTuplingUtils as ptu
from context import pwd

evaluator = ptu.boolean.eval.BooleanEvaluator
rb = ptu.io.read_branch


class ArithmeticTest(unittest.TestCase):
    ntp = uproot.open(osp.join(pwd, '../samples/sample.root'))
    tree = 'TupleB0/DecayTree'
    known_symb = {
        'pi': 3.14,
        'e': 2.72,
        'g': 9.8,
    }
    exe = evaluator(ntp, tree, known_symb=known_symb)

    ########
    # atom #
    ########

    def test_num(self):
        self.assertEqual(self.exe.eval('-1'), -1)
        self.assertEqual(type(self.exe.eval('-1')), int)
        self.assertEqual(type(self.exe.eval('1.85')), float)

    def test_bool(self):
        self.assertTrue(self.exe.eval('true'))
        self.assertFalse(self.exe.eval('false'))

    def test_known_syml(self):
        self.assertEqual(self.exe.eval('g'), 9.8)
        self.assertEqual(self.exe.eval('-pi'), -3.14)

    def test_var(self):
        self.assertTrue(np.array_equal(
            self.exe.eval('Y_ISOLATION_BDT3'),
            rb(self.ntp, self.tree, 'Y_ISOLATION_BDT3')))

    ###########
    # product #
    ###########

    def test_mul(self):
        self.assertEqual(self.exe.eval('3*pi'), 3*3.14)

    def test_div(self):
        self.assertEqual(self.exe.eval('g/pi'), 9.8/3.14)

    def test_mul_div(self):
        self.assertEqual(self.exe.eval('3*(g/pi)'), 3*(9.8/3.14))

    #######
    # sum #
    #######

    def test_add(self):
        self.assertEqual(self.exe.eval('-3+pi'), -3+3.14)

    def test_sub(self):
        self.assertEqual(self.exe.eval('3-pi'), 3-3.14)

    def test_arith(self):
        self.assertEqual(self.exe.eval('3*(pi+3)/(g-2)'), 3*(3.14+3)/(9.8-2))


class BooleanTest(unittest.TestCase):
    ntp = uproot.open(osp.join(pwd, '../samples/sample.root'))
    tree = 'TupleB0/DecayTree'
    exe = evaluator(ntp, tree)

    ##############
    # complement #
    ##############

    def test_comp(self):
        self.assertFalse(self.exe.eval('!true'))

    ##############
    # comparison #
    ##############

    def test_eq(self):
        self.assertTrue(self.exe.eval('!true == false'))

    def test_neq(self):
        self.assertFalse(self.exe.eval('!True != False'))  # case doesn't matter

    def test_gt(self):
        self.assertTrue(self.exe.eval('3 > 2'))

    def test_gte(self):
        self.assertTrue(self.exe.eval('pi >= 3'))

    ###########
    # boolean #
    ###########

    def test_andop(self):
        self.assertTrue(self.exe.eval('true & true'))
        self.assertFalse(self.exe.eval('true & false'))
        self.assertFalse(self.exe.eval('false & true'))
        self.assertFalse(self.exe.eval('false & false'))

    def test_orop(self):
        self.assertTrue(self.exe.eval('true | true'))
        self.assertTrue(self.exe.eval('true | false'))
        self.assertTrue(self.exe.eval('false | true'))
        self.assertFalse(self.exe.eval('false | false'))

    def test_andop_orop(self):
        self.assertTrue(self.exe.eval('false | true & true'))
        self.assertFalse(self.exe.eval('false | false & true'))

    def test_bool(self):
        self.assertTrue(self.exe.eval('(pi > e) & true | false'))
        self.assertFalse(self.exe.eval('(pi <= e) & (true | false)'))


class FunctionCallTest(unittest.TestCase):
    ntp = uproot.open(osp.join(pwd, '../samples/sample.root'))
    tree = 'TupleB0/DecayTree'
    exe = evaluator(ntp, tree)

    def test_func_call_zero_arg(self):
        self.assertEqual(self.exe.eval('(ONE())'), 1)

    def test_func_call_one_arg(self):
        self.assertEqual(self.exe.eval('abs(-1)'), 1)

    def test_func_call_one_arg_with_arithmetic(self):
        self.assertEqual(self.exe.eval('abs(-1-3*8)'), 25)

    def test_func_call_two_args(self):
        self.assertTrue(self.exe.eval('GT(pi, e)'))

    def test_func_call_nested(self):
        self.assertEqual(self.exe.eval('abs(-pi+ONE())'), np.abs(-np.pi+1))

    def test_func_call_nested_boolean_op(self):
        self.assertTrue(self.exe.eval('abs(abs(pi+2)*e) > abs(ONE()+e)'))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Sun May 09, 2021 at 02:52 AM +0200

import numpy as np

ARRAY_TYPE = 'np'


def read_branch(ntp, tree, branch, idx=None):
    data = ntp[tree][branch].array(library=ARRAY_TYPE)

    return data if not idx else data[idx]


def read_branches_dict(ntp, tree, branches):
    return ntp[tree].arrays(branches, library=ARRAY_TYPE)


def read_branches(ntp, tree, branches, idx=None, transpose=False):
    data = list(ntp[tree].arrays(branches, library=ARRAY_TYPE).values())

    if idx is not None:
        data = [d[idx] for d in data]

    return np.column_stack(data) if transpose else data



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Fri May 01, 2020 at 07:03 PM +0800

import unittest
import os.path as osp

from context import pyTuplingUtils as ptu
from context import pwd

rule = ptu.cutflow.CutflowRule
cfg = ptu.cutflow.CutflowGen
evaluator = ptu.boolean.eval.BooleanEvaluator


class CutflowTest(unittest.TestCase):
    ntp_path = osp.join(pwd, '../samples/sample.root')
    tree = 'TupleB0/DecayTree'
    exe = evaluator(ntp_path, tree)

    def test_cutflow_canonical(self):
        rules = [
            rule('muplus_L0Global_TIS & (Y_L0Global_TIS | Dst_2010_minus_L0HadronDecision_TOS)', key='L0'),
            rule('Kplus_Hlt1Phys_Dec', key='Hlt1'),
            rule('D0_Hlt2XcMuXForTauB2XcMuDecision_Dec', key='Hlt2'),
            rule('muplus_isMuon & muplus_PIDmu > 2', r'$\mu$ PID'),
            rule('Y_ISOLATION_BDT < 0.15', r'$\text{IsoBDT}_{\Upsilon(\text{4s})} < 0.15$'),
            rule('Y_M < 5280', r'$m_{\Upsilon(\text{4s})} < 5280$'),
        ]

        result = cfg(self.ntp_path, self.tree, rules, 2333).do()

        self.assertEqual(result['L0']['input'], 2333)
        self.assertEqual(result['L0']['output'], 176)
        self.assertEqual(result['muplus_isMuon & muplus_PIDmu > 2']['input'], 176)
        self.assertEqual(result['muplus_isMuon & muplus_PIDmu > 2']['output'], 167)
        self.assertEqual(result['muplus_isMuon & muplus_PIDmu > 2']['name'], r'$\mu$ PID')
        self.assertEqual(result['Y_M < 5280']['input'], 119)
        self.assertEqual(result['Y_M < 5280']['output'], 119)

    def test_cutflow_exotic(self):
        rules = [
            rule('Dst_2010_minus_L0HadronDecision_TOS'),
            rule('Dst_2010_minus_L0HadronDecision_TOS & Y_L0MuonDecision_TIS', compare_to=0, explicit=True),
            rule('Dst_2010_minus_L0HadronDecision_TOS & Y_L0ElectronDecision_TIS', compare_to=0, explicit=True),
        ]

        result = cfg(self.ntp_path, self.tree, rules, 2333).do()

        self.assertEqual(result['Dst_2010_minus_L0HadronDecision_TOS']['input'], 2333)
        self.assertEqual(result['Dst_2010_minus_L0HadronDecision_TOS']['output'], 85)
        self.assertEqual(result['Dst_2010_minus_L0HadronDecision_TOS & Y_L0MuonDecision_TIS']['input'], 85)
        self.assertEqual(result['Dst_2010_minus_L0HadronDecision_TOS & Y_L0MuonDecision_TIS']['output'], 8)
        self.assertEqual(result['Dst_2010_minus_L0HadronDecision_TOS & Y_L0ElectronDecision_TIS']['input'], 85)
        self.assertEqual(result['Dst_2010_minus_L0HadronDecision_TOS & Y_L0ElectronDecision_TIS']['output'], 10)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Sun Jun 13, 2021 at 12:38 AM +0200

import unittest

from context import pyTuplingUtils as ptu
from context import pwd


class SplitNtpTreeTest(unittest.TestCase):
    def test_split(self):
        ntp_tree = 'a/b/c/d.root/TupleB0/DecayTree'
        ntp, tree = ptu.argparse.split_ntp_tree(ntp_tree)

        assert ntp == 'a/b/c/d.root'
        assert tree == 'TupleB0/DecayTree'



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Thu Jun 10, 2021 at 02:31 AM +0200

__name__ = 'pyTuplingUtils'
__version__ = '0.2.11'

import pyTuplingUtils.boolean
import pyTuplingUtils.io
import pyTuplingUtils.argparse
import pyTuplingUtils.plot
import pyTuplingUtils.utils
import pyTuplingUtils.cutflow



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Wed Jun 17, 2020 at 03:16 AM +0800

from lark import Lark


boolean_grammar = '''
    ?start: boolor

    ?boolor: booland
        | boolor "|" booland -> orop

    ?booland: cond  // '&' binds tigher than '|'
        | booland "&" cond   -> andop

    ?cond: expr
        | cond "==" expr     -> eq
        | cond "!=" expr     -> neq
        | cond ">"  expr     -> gt
        | cond ">=" expr     -> gte
        | cond "<"  expr     -> lt
        | cond "<=" expr     -> lte

    ?expr: sum
        | "!" sum            -> comp  // logical complement

    ?sum: product
        | sum "+" product    -> add
        | sum "-" product    -> sub

    ?product: molecule
        | product "*" molecule   -> mul
        | product "/" molecule   -> div

    ?molecule: NAME "(" [arglist] ")" -> func_call
        | atom

    ?atom: NUMBER            -> num
        | "-" atom           -> neg
        | BOOL               -> bool
        | NAME               -> var
        | "(" boolor ")"

    arglist: (boolor ",")* (boolor [","])

    %import common.SIGNED_NUMBER -> NUMBER
    %import common.WS_INLINE
    %import common.CNAME -> NAME

    %ignore WS_INLINE

    BOOL.2: "True" | "False" | "true" | "false"  // These keywords have higher priority
'''

boolean_parser = Lark(boolean_grammar, parser='lalr')



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Tue Nov 23, 2021 at 03:51 AM +0100

import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt

from functools import wraps
from inspect import getfullargspec


###################
# General helpers #
###################

def filter_kwargs_func(kwargs, func):
    kw_func = getfullargspec(func).args

    kw_known = {k: kwargs[k] for k in kw_func if k in kwargs}
    kw_rest = {k: kwargs[k] for k in kwargs if k not in kw_known}

    return kw_known, kw_rest


def decorate_output(f, tight_layout=True, pad=0.):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)
        output = result[0]
        data = result[1:]

        if not output:
            return data

        fig = data[0]
        if tight_layout:
            plt.tight_layout(pad=pad)
        fig.savefig(output)
        fig.clf()
        plt.close(fig)

    return wrapper


def decorate_ax_style(f, *args, **kwargs):
    @wraps(f)
    def wrapper(*args, **kwargs):
        kw_known, kw_rest = filter_kwargs_func(kwargs, f)

        result = f(*args, **kw_known)
        result.update(kw_rest)

        return result

    return wrapper


################
# Plot helpers #
################

def tick_formatter_short(x, p):
    X = str(x)
    if len(X) > 4:
        return '{:2g}'.format(x)
    return x


def plot_prepare(figure=None, axis=None, title=None,
                 xtick_formatter=None, ytick_formatter=None,
                 xlabel=None, ylabel=None,
                 xlim=None, ylim=None,
                 xscale='linear', yscale='linear',
                 show_legend=True,
                 legend_add_args={'numpoints': 1, 'loc': 'best'}):
    fig = plt.figure() if not figure else figure

    if not axis:
        ax = fig.add_subplot()
        ax.set_xscale(xscale)
        ax.set_yscale(yscale)
    else:
        ax = axis

    if title:
        ax.set_title(title)

    if xtick_formatter:
        ax.get_xaxis().set_major_formatter(
            mp.ticker.FuncFormatter(xtick_formatter)
        )
    if ytick_formatter:
        ax.get_yaxis().set_major_formatter(
            mp.ticker.FuncFormatter(ytick_formatter)
        )

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    if xlim:
        ax.set_xlim(xlim)
    if ylim:
        ax.set_ylim(ylim)

    if show_legend:
        legend = lambda: ax.legend(**legend_add_args)
    else:
        legend = lambda: True

    return fig, ax, legend


def get_ytick_position_in_ax_coordinate(ax, scale):
    y_min, y_max = ax.get_ylim()
    raw_ticks = ax.get_yticks(minor=False)

    if scale == 'linear':
        ticks = [(tick - y_min)/(y_max - y_min) for tick in raw_ticks]
    elif scale == 'log':
        crd = np.vstack((np.zeros_like(raw_ticks), raw_ticks)).T
        ticks = ax.transAxes.inverted().transform(
            ax.transData.transform(crd))[:, 1]
    else:
        raise ValueError('Unknown axis scale: {}'.format(scale))

    return y_min, y_max, ticks


def ensure_no_majortick_on_topmost(ax, scale, thresh=0.9, ratio=0.19,
                                   verbose=False):
    y_min, y_max, ticks = get_ytick_position_in_ax_coordinate(ax, scale)

    if verbose:
        print('y range is: {}, {}'.format(y_min, y_max))
        print('y major ticks are: {}'.format(','.join([str(i) for i in ticks])))

    max_majortick = ticks[-2] if ticks[-1] > 1 else ticks[-1]
    if verbose:
        print('Max y tick is at: {}'.format(max_majortick))

    if max_majortick >= thresh:
        rel_tick_gap = ticks[1] - ticks[0]
        abs_tick_gap = rel_tick_gap * (y_max-y_min)
        padding = ratio * abs_tick_gap

        ax.set_ylim((y_min, y_max+padding))
        if verbose:
            print('Added y pad. Now y max is: {}'.format(y_max+padding))


################
# Data helpers #
################

def convert_bins_to_central_pos(bins):
    return bins[:-1] + (np.diff(bins)/2)


######################
# Plot style helpers #
######################

@decorate_ax_style
def ax_add_args_histo(label, color='blue', edgecolor='none'):
    if not edgecolor:
        edgecolor = color

    return {
        'color': color,
        'edgecolor': 'none',
        'label': label
    }


@decorate_ax_style
def ax_add_args_errorbar(label, color, yerr=None, marker='o',
                         markeredgecolor='none', ls='none'):
    return {
        'label': label,
        'ls': ls,
        'color': color,
        'marker': marker,
        'markeredgecolor': markeredgecolor,
        'yerr': yerr
    }


@decorate_ax_style
def ax_add_args_step(label, color, where='post'):
    return {
        'label': label,
        'color': color,
        'where': where
    }


@decorate_ax_style
def ax_add_args_fill(color, alpha=0.5, linewidth=0., step='post'):
    return {
        'color': color,
        'alpha': alpha,
        'linewidth': linewidth,
        'step': step
    }


@decorate_ax_style
def ax_add_args_hlines(label, color, linestyles='solid'):
    return {
        'label': label,
        'colors': color,
        'linestyles': linestyles
    }


ax_add_args_vlines = ax_add_args_hlines


################
# Simple plots #
################

@decorate_output
def plot_hexbin(x, y, gridsize, hexbin_add_args,
                output=None, cmap='inferno', bins='log', colorbar_label=None,
                **kwargs):
    fig, ax, legend = plot_prepare(**kwargs)
    hb = ax.hexbin(x, y,
                   gridsize=gridsize, cmap=cmap, bins=bins, **hexbin_add_args)
    cb = fig.colorbar(hb, ax=ax)

    if colorbar_label:
        cb.set_label(colorbar_label)

    legend()
    return output, fig, ax


@decorate_output
def plot_histo(bins, histo, histo_add_args,
               output=None, xtick_formatter=tick_formatter_short,
               **kwargs):
    fig, ax, legend = plot_prepare(xtick_formatter=xtick_formatter, **kwargs)
    mid_bins = convert_bins_to_central_pos(bins)
    ax.bar(mid_bins, histo, width=np.diff(bins), **histo_add_args)

    legend()
    return output, fig, ax


@decorate_output
def plot_errorbar(x, y, errorbar_add_args,
                  output=None, convert_x=True,
                  **kwargs):
    if convert_x:
        x = convert_bins_to_central_pos(x)

    fig, ax, legend = plot_prepare(**kwargs)
    ax.errorbar(x, y, **errorbar_add_args)

    legend()
    return output, fig, ax


@decorate_output
def plot_step(x, y, step_add_args,
              output=None,
              **kwargs):
    y = np.append(y, y[-1])

    fig, ax, legend = plot_prepare(**kwargs)
    ax.step(x, y, **step_add_args)

    legend()
    return output, fig, ax


@decorate_output
def plot_fill(x, y_range, fill_add_args,
              output=None,
              **kwargs):
    ymin, ymax = y_range
    ymin = np.append(ymin, ymin[-1])
    ymax = np.append(ymax, ymax[-1])

    fig, ax, legend = plot_prepare(**kwargs)
    ax.fill_between(x, ymin, ymax, **fill_add_args)

    legend()
    return output, fig, ax


@decorate_output
def plot_hlines(x, y, hlines_add_args,
                output=None,
                **kwargs):
    xmin = x[:-1]
    xmax = x[1:]

    fig, ax, legend = plot_prepare(**kwargs)
    ax.hlines(y, xmin, xmax, **hlines_add_args)

    legend()
    return output, fig, ax


@decorate_output
def plot_vlines(x, y, vlines_add_args,
                output=None, convert_x=True,
                **kwargs):
    if convert_x:
        x = convert_bins_to_central_pos(x)

    ymin = y[0]
    ymax = y[1]

    fig, ax, legend = plot_prepare(**kwargs)
    ax.vlines(x, ymin, ymax, **vlines_add_args)

    legend()
    return output, fig, ax


###################
# Composite plots #
###################

@decorate_output
def plot_top(top_plotters,
             output=None,
             legend_add_args={'numpoints': 1, 'loc': 'best'},
             **kwargs):
    fig, ax, _ = plot_prepare(**kwargs)

    for p in top_plotters:
        p(fig, ax)

    ax.legend(**legend_add_args)
    return output, fig, ax


def plot_top_bot(top_plotters, bot_plotters,
                 title=None,
                 xlabel=None, ax1_ylabel=None, ax2_ylabel=None,
                 ax1_yscale='linear', ax2_yscale='linear',
                 height_ratios=[3, 1],
                 legend_add_args={'numpoints': 1, 'loc': 'best'},
                 **kwargs):
    fig = plt.figure()
    fig.set_tight_layout({'pad': 0.})

    spec = fig.add_gridspec(ncols=1, nrows=2, height_ratios=height_ratios)
    spec.update(hspace=0.)

    ax1 = fig.add_subplot(spec[0, 0])
    ax1.set_yscale(ax1_yscale)

    if ax1_ylabel:
        ax1.set_ylabel(ax1_ylabel)
    if title:
        ax1.set_title(title)

    ax2 = fig.add_subplot(spec[1, 0], sharex=ax1)
    ax2.set_yscale(ax2_yscale)

    if xlabel:
        ax2.set_xlabel(xlabel)
    if ax2_ylabel:
        ax2.set_ylabel(ax2_ylabel)

    for p in top_plotters:
        p(fig, ax1, **kwargs)

    for p in bot_plotters:
        p(fig, ax2, **kwargs)

    # Always show legend for top
    ax1.legend(**legend_add_args)

    # Remove the horizontal labels for the top plot
    for tick in ax1.xaxis.get_major_ticks():
        tick.label1.set_visible(False)

    # No offset (like +1 on top of the axis)
    if ax2_yscale == 'linear':
        ax2.ticklabel_format(axis='y', useOffset=False)

    # Add some padding if the TOPMOST ylabel of bot is too close to the top
    ensure_no_majortick_on_topmost(ax2, ax2_yscale)

    # Align y labels
    fig.align_ylabels()

    return fig, ax1, ax2



#!/usr/bin/env python3
#
# Authorop: Yipeng Sun
# License: BSD 2-clause
# Last Change: Thu Jun 18, 2020 at 02:47 AM +0800

import unittest

from context import pyTuplingUtils as ptu

parser = ptu.boolean.syntax.boolean_parser.parse


class ArithmeticTest(unittest.TestCase):
    def test_var(self):
        self.assertEqual(
            parser('a').pretty(),
            "var\ta\n"
        )
        self.assertEqual(
            parser('-a').pretty(),
            "neg\n"
            "  var\ta\n"
        )
        self.assertEqual(
            parser('a1').pretty(),
            "var\ta1\n"
        )
        self.assertEqual(
            parser('a_1b').pretty(),
            "var\ta_1b\n"
        )

    def test_num(self):
        self.assertEqual(
            parser('1').pretty(),
            "num\t1\n"
        )
        self.assertEqual(
            parser('+1').pretty(),
            "num\t+1\n"
        )

    def test_negative_num(self):
        self.assertEqual(
            parser('-1.6').pretty(),
            "num\t-1.6\n"
        )
        self.assertEqual(
            parser('+1').pretty(),
            "num\t+1\n"
        )

    def test_add(self):
        self.assertEqual(
            parser('-1 +2.3').pretty(),
            "add\n"
            "  num\t-1\n"
            "  num\t2.3\n"
        )

    def test_add_sub(self):
        self.assertEqual(
            parser('-1 +2.3 - 10').pretty(),
            "sub\n"
            "  add\n"
            "    num\t-1\n"
            "    num\t2.3\n"
            "  num\t10\n"
        )

    def test_add_mul(self):
        self.assertEqual(
            parser('-1 +2.3 * 10').pretty(),
            "add\n"
            "  num\t-1\n"
            "  mul\n"
            "    num\t2.3\n"
            "    num\t10\n"
        )

    def test_add_mul_par(self):
        self.assertEqual(
            parser('-(1 +2.3) * 10').pretty(),
            "mul\n"
            "  neg\n"
            "    add\n"
            "      num\t1\n"
            "      num\t2.3\n"
            "  num\t10\n"
        )


class BooleanTest(unittest.TestCase):
    def test_comp(self):
        self.assertEqual(
            parser('!(-a_2 +2.3) ').pretty(),
            "comp\n"
            "  add\n"
            "    neg\n"
            "      var\ta_2\n"
            "    num\t2.3\n"
        )

    def test_eq(self):
        self.assertEqual(
            parser('a == b').pretty(),
            "eq\n"
            "  var\ta\n"
            "  var\tb\n"
        )
        self.assertEqual(
            parser('a == 1').pretty(),
            "eq\n"
            "  var\ta\n"
            "  num\t1\n"
        )
        self.assertEqual(
            parser('a <= -1+x').pretty(),
            "lte\n"
            "  var\ta\n"
            "  add\n"
            "    num\t-1\n"
            "    var\tx\n"
        )

    def test_bool(self):
        self.assertEqual(
            parser('a & 1').pretty(),
            "andop\n"
            "  var\ta\n"
            "  num\t1\n"
        )
        self.assertEqual(
            parser('True | False').pretty(),
            "orop\n"
            "  bool\tTrue\n"
            "  bool\tFalse\n"
        )
        self.assertEqual(
            parser('True | False & True').pretty(),
            "orop\n"
            "  bool\tTrue\n"
            "  andop\n"
            "    bool\tFalse\n"
            "    bool\tTrue\n"
        )
        self.assertEqual(
            parser('(True | False) & !True | false').pretty(),
            "orop\n"
            "  andop\n"
            "    orop\n"
            "      bool\tTrue\n"
            "      bool\tFalse\n"
            "    comp\n"
            "      bool\tTrue\n"
            "  bool\tfalse\n"
        )

    def test_comb(self):
        self.assertEqual(
            parser('a >= !(-1+x)*3').pretty(),
            "gte\n"
            "  var\ta\n"
            "  comp\n"
            "    mul\n"
            "      add\n"
            "        num\t-1\n"
            "        var\tx\n"
            "      num\t3\n"
        )
        self.assertEqual(
            parser('a >= !(-1+x)*3 | x<8 & y != -(z+3)').pretty(),
            "orop\n"
            "  gte\n"
            "    var\ta\n"
            "    comp\n"
            "      mul\n"
            "        add\n"
            "          num\t-1\n"
            "          var\tx\n"
            "        num\t3\n"
            "  andop\n"
            "    lt\n"
            "      var\tx\n"
            "      num\t8\n"
            "    neq\n"
            "      var\ty\n"
            "      neg\n"
            "        add\n"
            "          var\tz\n"
            "          num\t3\n"
        )
        self.assertEqual(
            parser('a >= !(-1+x)*3 | x<8 & y != -(z+3)').pretty(),
            parser('a >= !(-1+x)*3 | (x<8 & y != -(z+3))').pretty()
        )


class FunctionCallTest(unittest.TestCase):
    def test_func_call_zero_arg(self):
        self.assertEqual(
            parser('(some_func0())').pretty(),
            "func_call\tsome_func0\n"
        )

    def test_func_call_one_arg(self):
        self.assertEqual(
            parser('some_func1(arg1)').pretty(),
            "func_call\n"
            "  some_func1\n"
            "  arglist\n"
            "    var\targ1\n"
        )

    def test_func_call_two_args(self):
        self.assertEqual(
            parser('some_func2(arg1, arg2)').pretty(),
            "func_call\n"
            "  some_func2\n"
            "  arglist\n"
            "    var\targ1\n"
            "    var\targ2\n"
        )

    def test_func_call_arithmetic(self):
        self.assertEqual(
            parser('arith_func((arg1+2)*val3, arg2)').pretty(),
            "func_call\n"
            "  arith_func\n"
            "  arglist\n"
            "    mul\n"
            "      add\n"
            "        var\targ1\n"
            "        num\t2\n"
            "      var\tval3\n"
            "    var\targ2\n"
        )

    def test_func_call_nested(self):
        self.assertEqual(
            parser('arith_func(inner(arg1+2)*val3, arg2)').pretty(),
            "func_call\n"
            "  arith_func\n"
            "  arglist\n"
            "    mul\n"
            "      func_call\n"
            "        inner\n"
            "        arglist\n"
            "          add\n"
            "            var\targ1\n"
            "            num\t2\n"
            "      var\tval3\n"
            "    var\targ2\n"
        )

    def test_func_call_nested_boolean_op(self):
        self.assertEqual(
            parser('arith_func(inner(arg1+2)*val3, arg2) > stuff(a)').pretty(),
            "gt\n"
            "  func_call\n"
            "    arith_func\n"
            "    arglist\n"
            "      mul\n"
            "        func_call\n"
            "          inner\n"
            "          arglist\n"
            "            add\n"
            "              var\targ1\n"
            "              num\t2\n"
            "        var\tval3\n"
            "      var\targ2\n"
            "  func_call\n"
            "    stuff\n"
            "    arglist\n"
            "      var\ta\n"
        )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Fri May 01, 2020 at 04:30 AM +0800

import sys

from os.path import abspath, join, dirname

pwd = abspath(join(dirname(__file__)))

sys.path.insert(0, join(pwd, '..'))

import pyTuplingUtils



# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Thu Jul 29, 2021 at 02:39 AM +0200

import setuptools
import codecs
import os.path

from distutils.core import setup


###########
# Helpers #
###########

with open('README.md', 'r') as ld:
    long_description = ld.read()


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


#########
# Setup #
#########

setup(
    name='pyTuplingUtils',
    version=get_version('pyTuplingUtils/__init__.py'),
    author='Yipeng Sun',
    author_email='syp@umd.edu',
    description='Utilities for ntuples, such as plotting and simple debugging, with the help of uproot',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/umd-lhcb/pyTuplingUtils',
    packages=setuptools.find_packages(),
    scripts=['bin/uidcommon', 'bin/uiddump', 'bin/plotbr', 'bin/plotbrdiff'],
    include_package_data=True,
    install_requires=[
        'uproot',
        'lz4',
        'matplotlib',
        'lark-parser',
        'mplhep'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent'
    ]
)



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Wed Jun 24, 2020 at 09:08 PM +0800

from lark import Transformer, v_args

from numpy import logical_and as AND, logical_or as OR, logical_not as NOT
from pyTuplingUtils.io import read_branches_dict
from pyTuplingUtils.boolean.syntax import boolean_parser
from pyTuplingUtils.boolean.const import KNOWN_SYMB, KNOWN_FUNC


class TransForTupling(Transformer):
    def __init__(self, ntp, tree,
                 known_symb=KNOWN_SYMB, known_func=KNOWN_FUNC):
        self.ntp = ntp
        self.tree = tree
        self.cache = {}
        self.cache.update(known_symb)
        self.known_symb = known_symb
        self.known_func = known_func

    ########
    # atom #
    ########

    @v_args(inline=True)
    def num(self, val):
        try:
            return int(val)
        except ValueError:
            return float(val)

    @v_args(inline=True)
    def bool(self, val):
        return True if val.lower() == 'true' else False

    @v_args(inline=True)
    def var(self, val):
        return self.cache[val.value]

    @v_args(inline=True)
    def neg(self, val):
        return -val

    ###########
    # product #
    ###########

    @v_args(inline=True)
    def mul(self, arg1, arg2):
        return arg1 * arg2

    @v_args(inline=True)
    def div(self, arg1, arg2):
        return arg1 / arg2

    #######
    # sum #
    #######

    @v_args(inline=True)
    def add(self, arg1, arg2):
        return arg1 + arg2

    @v_args(inline=True)
    def sub(self, arg1, arg2):
        return arg1 - arg2

    ##############
    # complement #
    ##############

    @v_args(inline=True)
    def comp(self, cond):
        return NOT(cond)

    ##############
    # comparison #
    ##############

    @v_args(inline=True)
    def eq(self, lhs, rhs):
        return lhs == rhs

    @v_args(inline=True)
    def neq(self, lhs, rhs):
        return lhs != rhs

    @v_args(inline=True)
    def gt(self, lhs, rhs):
        return lhs > rhs

    @v_args(inline=True)
    def gte(self, lhs, rhs):
        return lhs >= rhs

    @v_args(inline=True)
    def lt(self, lhs, rhs):
        return lhs < rhs

    @v_args(inline=True)
    def lte(self, lhs, rhs):
        return lhs <= rhs

    ###########
    # boolean #
    ###########

    @v_args(inline=True)
    def andop(self, cond1, cond2):
        return AND(cond1, cond2)

    @v_args(inline=True)
    def orop(self, cond1, cond2):
        return OR(cond1, cond2)

    #################
    # Function call #
    #################

    @v_args(inline=True)
    def func_call(self, func_name, arglist=None):
        if arglist is not None:
            return self.known_func[str(func_name)](*arglist.children)
        else:
            return self.known_func[str(func_name)]()


class BooleanEvaluator(object):
    def __init__(self, *args, transformer=TransForTupling, **kwargs):
        self.parser = boolean_parser.parse
        self.transformer = transformer(*args, **kwargs)

    def parse(self, s):
        return self.parser(s)

    def eval(self, s):
        tree = self.parse(s)

        # Load all variables in the expression in batch
        vars_to_load = [str(n.children[0]) for n in tree.find_data('var')
                        if str(n.children[0]) not in
                        self.transformer.known_symb.keys()]
        self.transformer.cache.update(read_branches_dict(
            self.transformer.ntp, self.transformer.tree, vars_to_load))

        return self.transformer.transform(tree)



#!/usr/bin/env python3
#
# Author: Yipeng Sun
# License: BSD 2-clause
# Last Change: Thu Jul 30, 2020 at 04:03 AM +0800

import numpy as np


KNOWN_SYMB = {
    # Mathematical constants
    'pi': np.pi,
    'e': np.e,
    # Units
    'MeV': 1.0,
    'GeV': 1000.0,
    # Particle masses
    # B mesons
    'PDG_M_B0': 5279.64,
    # D mesons
    'PDG_M_Dst': 2010.26,
    'PDG_M_D0': 1864.83,
}

KNOWN_FUNC = {
    # NOTE: Convention is lowercase for functions that ROOT can parse, and uppercase otherwise
    'abs': np.abs,
    'log': np.log,
    'sin': np.sin,
    'ONE': lambda: 1,
    'LOG10pp': lambda p1x, p1y, p1z, p2x, p2y, p2z: np.log10(1-(p1x*p2x + p1y*p2y + p1z*p2z)/np.sqrt(p1x*p1x+p1y*p1y+p1z*p1z)/np.sqrt(p2x*p2x+p2y*p2y+p2z*p2z)),
    'ETA': lambda p, pz: np.log((p+pz)/(p-pz))/2.,
    'NORM2': lambda x, y: np.sqrt(x*x + y*y),
    'GT': lambda x, y: x > y,
    'LT': lambda x, y: x < y,
}
